export namespace MeterbusLib {
    export class InvalidTelegramError extends Error {
    }

    export class InvalidStartCharError extends InvalidTelegramError {
    }

    export class InvalidLengthError extends InvalidTelegramError {        
    }

    export class InvalidChecksumError extends InvalidTelegramError {        
    }

    export class InvalidStopCharError extends InvalidTelegramError {        
    }

    export class InvalidSecondLengthError extends InvalidTelegramError {
    }

    export class PayloadTooShortError extends InvalidTelegramError {
    }

    export abstract class Frame {
        protected _telegram : number[]
        protected _startCharacter : number
        protected _frameLength : number
        protected _firstPayload : number
        protected _payloadLength : number

        constructor(startCharacter : number, frameLength : number, 
            firstPayload : number, telegram : number[]) {
            this._startCharacter = startCharacter
            this._frameLength = frameLength
            this._firstPayload = firstPayload
            this._telegram = telegram
            this._payloadLength = this._frameLength - 6

        }

        verifyChecksumAndEnd() {
            if (this._firstPayload != 0) {
                if (this._telegram[this._telegram.length - 1] != 0x16) {
                    throw new InvalidStopCharError()
                }
                let checksum : number = 0
                this._telegram.slice(this._firstPayload, this._telegram.length -2).forEach((val) => {
                    checksum += val
                })
                // console.log(`calc. checksum ${checksum}, ${checksum & 0xff}`)
                // console.log(`found checksum ${this._telegram[this._telegram.length - 2]}`)
                if ((checksum & 0xff) != this._telegram[this._telegram.length - 2]) {
                    throw new InvalidChecksumError()
                }
            }
        }
    
        abstract parse2()

        parse() {
            if (this._telegram[0] != this._startCharacter) {
                throw new InvalidStartCharError()
            }
            if (this._telegram.length != this._frameLength) {
                throw new InvalidLengthError()
            }
            this.verifyChecksumAndEnd()
            this.parse2()
        }
        
    }

    export class ControlFrame extends Frame {
        private _cField : number
        private _address : number
        private _ciField : number

        get cField() : number {
            return this._cField
        }

        get address() : number {
            return this._address
        }

        get ciField() : number {
            return this._ciField
        }

        constructor(telegram : number[]) {
            super(0x68, (telegram[1] + 6), 4, telegram)
        }

        parse2() {
            if (this._telegram[2] != this._payloadLength) {
                throw new InvalidSecondLengthError()
            }
            if (this._payloadLength < 3) {
                throw new PayloadTooShortError()
            }

            this._cField = this._telegram[4]
            this._address = this._telegram[5]            
            this._ciField = this._telegram[6]            
        }
    }

    export class ShortFrame extends Frame {
        private _cField : number
        private _address : number

        get cField() : number {
            return this._cField
        }

        get address() : number {
            return this._address
        }

        constructor(telegram : number[]) {
            super(0x10, 5, 1, telegram)
        }

        parse2() {
            this._cField = this._telegram[1]
            this._address = this._telegram[2]            
        }
    }

    export class LongFrame extends ControlFrame {
        constructor(telegram : number[]) {
            super(telegram)
        }

        parse2() {
            super.parse2()
        }
    }

    export class SingleCharFrame extends Frame {
        constructor(telegram : number[]) {
            super(0xe5, 1, 0, telegram)
        }

        parse2() {
            
        }
    }

    export class Telegram {
        private _hexString : string
        private _telegram : number[]
        private _frame : Frame

        constructor() {
        }
    
        get hexString() : string {
            return this._hexString
        }

        get telegram() : number[] {
            return this._telegram
        }

        get frame() : Frame {
            return this._frame
        }

        fromHexString(hexString : string) {
            this._hexString = hexString
            this._telegram = this._hexString.split(' ').map<number>((val) => {
                return parseInt(val, 16)
            })
        }

        parse() {
            if (this._telegram[0] === 0x68 && this._telegram[1] === 0x03) {
                this._frame = new ControlFrame(this._telegram)
            } else if (this._telegram[0] === 0x68) {
                this._frame = new LongFrame(this._telegram)
            } else if (this._telegram[0] === 0x10) {
                this._frame = new ShortFrame(this._telegram)
            } else if (this._telegram[0] === 0xe5) {
                this._frame = new SingleCharFrame(this._telegram)
            } else {
                throw new InvalidStartCharError()
            }

            this._frame.parse()
        }
    }
}